import datetime
import uuid as uuid_package
import time
from typing import Any, List, Dict, Union, cast

import grpc  # type: ignore
from google.protobuf.struct_pb2 import Struct

from weaviate.collections.classes.batch import (
    ErrorObject,
    _BatchObject,
    BatchObjectReturn,
)
from weaviate.collections.classes.internal import _Reference
from weaviate.collections.grpc.shared import _BaseGRPC
from weaviate.exceptions import WeaviateQueryException, WeaviateInsertInvalidPropertyError
from weaviate.util import _datetime_to_string
from weaviate.proto.v1 import batch_pb2, base_pb2


class _BatchGRPC(_BaseGRPC):
    """This class is used to insert multiple objects into Weaviate using the gRPC API.

    It is used within the `_Data` and `_Batch` classes hence the necessary generalities
    and abstractions so as not to couple to strongly to either use-case.
    """

    def objects(self, objects: List[_BatchObject]) -> BatchObjectReturn:
        """Insert multiple objects into Weaviate through the gRPC API.

        Parameters:
            `objects`
                A list of `WeaviateObject` containing the data of the objects to be inserted. The class name must be
                provided for each object, and the UUID is optional. If no UUID is provided, one will be generated for each object.
                The UUIDs of the inserted objects will be returned in the `uuids` attribute of the returned `_BatchReturn` object.
                The UUIDs of the objects that failed to be inserted will be returned in the `errors` attribute of the returned `_BatchReturn` object.
            `tenant`
                The tenant to be used for this batch operation
        """
        weaviate_objs: List[batch_pb2.BatchObject] = [
            batch_pb2.BatchObject(
                collection=obj.class_name,
                vector=obj.vector,
                uuid=str(obj.uuid) if obj.uuid is not None else str(uuid_package.uuid4()),
                properties=self.__translate_properties_from_python_to_grpc(obj.properties, False),
                tenant=obj.tenant,
            )
            for obj in objects
        ]

        start = time.time()
        errors = self.__send_batch(weaviate_objs)
        elapsed_time = time.time() - start

        all_responses: List[Union[uuid_package.UUID, ErrorObject]] = cast(
            List[Union[uuid_package.UUID, ErrorObject]], list(range(len(weaviate_objs)))
        )
        return_success: Dict[int, uuid_package.UUID] = {}
        return_errors: Dict[int, ErrorObject] = {}

        for idx, obj in enumerate(weaviate_objs):
            if idx in errors:
                error = ErrorObject(errors[idx], objects[idx], original_uuid=objects[idx].uuid)
                return_errors[idx] = error
                all_responses[idx] = error
            else:
                success = uuid_package.UUID(obj.uuid)
                return_success[idx] = success
                all_responses[idx] = success

        return BatchObjectReturn(
            uuids=return_success,
            errors=return_errors,
            has_errors=len(errors) > 0,
            all_responses=all_responses,
            elapsed_seconds=elapsed_time,
        )

    def __send_batch(self, batch: List[batch_pb2.BatchObject]) -> Dict[int, str]:
        metadata = self._get_metadata()
        try:
            assert self._connection.grpc_stub is not None
            res: batch_pb2.BatchObjectsReply
            res, _ = self._connection.grpc_stub.BatchObjects.with_call(
                batch_pb2.BatchObjectsRequest(
                    objects=batch,
                    consistency_level=self._consistency_level,
                ),
                metadata=metadata,
            )

            objects: Dict[int, str] = {}
            for result in res.errors:
                objects[result.index] = result.error
            return objects
        except grpc.RpcError as e:
            raise WeaviateQueryException(e.details())

    def __translate_properties_from_python_to_grpc(
        self, data: Dict[str, Any], clean_props: bool
    ) -> batch_pb2.BatchObject.Properties:
        _validate_props(data, clean_props)

        multi_target: List[batch_pb2.BatchObject.MultiTargetRefProps] = []
        single_target: List[batch_pb2.BatchObject.SingleTargetRefProps] = []
        non_ref_properties: Struct = Struct()
        bool_arrays: List[base_pb2.BooleanArrayProperties] = []
        text_arrays: List[base_pb2.TextArrayProperties] = []
        int_arrays: List[base_pb2.IntArrayProperties] = []
        float_arrays: List[base_pb2.NumberArrayProperties] = []
        object_properties: List[base_pb2.ObjectProperties] = []
        object_array_properties: List[base_pb2.ObjectArrayProperties] = []
        for key, val in data.items():
            if isinstance(val, _Reference):
                if val.is_multi_target:
                    multi_target.append(
                        batch_pb2.BatchObject.MultiTargetRefProps(
                            uuids=val.uuids_str,
                            target_collection=val.target_collection,
                            prop_name=key,
                        )
                    )
                else:
                    single_target.append(
                        batch_pb2.BatchObject.SingleTargetRefProps(
                            uuids=val.uuids_str, prop_name=key
                        )
                    )
            elif isinstance(val, dict):
                parsed = self.__translate_properties_from_python_to_grpc(val, clean_props)
                object_properties.append(
                    base_pb2.ObjectProperties(
                        prop_name=key,
                        value=base_pb2.ObjectPropertiesValue(
                            non_ref_properties=parsed.non_ref_properties,
                            int_array_properties=parsed.int_array_properties,
                            text_array_properties=parsed.text_array_properties,
                            number_array_properties=parsed.number_array_properties,
                            boolean_array_properties=parsed.boolean_array_properties,
                            object_properties=parsed.object_properties,
                            object_array_properties=parsed.object_array_properties,
                        ),
                    )
                )
            elif isinstance(val, list) and isinstance(val[0], dict):
                val = cast(List[Dict[str, Any]], val)
                object_array_properties.append(
                    base_pb2.ObjectArrayProperties(
                        values=[
                            base_pb2.ObjectPropertiesValue(
                                non_ref_properties=parsed.non_ref_properties,
                                int_array_properties=parsed.int_array_properties,
                                text_array_properties=parsed.text_array_properties,
                                number_array_properties=parsed.number_array_properties,
                                boolean_array_properties=parsed.boolean_array_properties,
                                object_properties=parsed.object_properties,
                                object_array_properties=parsed.object_array_properties,
                            )
                            for v in val
                            if (
                                parsed := self.__translate_properties_from_python_to_grpc(
                                    v, clean_props
                                )
                            )
                        ],
                        prop_name=key,
                    )
                )
            elif isinstance(val, list) and isinstance(val[0], bool):
                bool_arrays.append(base_pb2.BooleanArrayProperties(prop_name=key, values=val))
            elif isinstance(val, list) and isinstance(val[0], str):
                text_arrays.append(base_pb2.TextArrayProperties(prop_name=key, values=val))
            elif isinstance(val, list) and isinstance(val[0], datetime.datetime):
                text_arrays.append(
                    base_pb2.TextArrayProperties(
                        prop_name=key, values=[_datetime_to_string(x) for x in val]
                    )
                )
            elif isinstance(val, list) and isinstance(val[0], uuid_package.UUID):
                text_arrays.append(
                    base_pb2.TextArrayProperties(prop_name=key, values=[str(x) for x in val])
                )
            elif isinstance(val, list) and isinstance(val[0], int):
                int_arrays.append(base_pb2.IntArrayProperties(prop_name=key, values=val))
            elif isinstance(val, list) and isinstance(val[0], float):
                float_arrays.append(base_pb2.NumberArrayProperties(prop_name=key, values=val))
            else:
                non_ref_properties.update({key: _serialize_primitive(val)})

        return batch_pb2.BatchObject.Properties(
            non_ref_properties=non_ref_properties,
            multi_target_ref_props=multi_target,
            single_target_ref_props=single_target,
            text_array_properties=text_arrays,
            number_array_properties=float_arrays,
            int_array_properties=int_arrays,
            boolean_array_properties=bool_arrays,
            object_properties=object_properties,
            object_array_properties=object_array_properties,
        )


def _validate_props(props: Dict[str, Any], clean_props: bool) -> None:
    should_throw = False
    if "id" in props:
        if clean_props:
            del props["id"]
        else:
            should_throw = True
    if "vector" in props:
        if clean_props:
            del props["vector"]
        else:
            should_throw = True
    if should_throw:
        raise WeaviateInsertInvalidPropertyError(props)


def _serialize_primitive(value: Any) -> Any:
    if isinstance(value, uuid_package.UUID):
        return str(value)
    if isinstance(value, datetime.datetime):
        return _datetime_to_string(value)
    if isinstance(value, list):
        return [_serialize_primitive(val) for val in value]
    return value



from weaviate.collections.classes.config import (
    Configure,
    ConfigureUpdate,
    DataType,
    Multi2VecField,
    Property,
    ReferenceProperty,
    ReferencePropertyMultiTarget,
    Tokenization,
    VectorDistance,
)
from weaviate.collections.classes.data import (
    DataObject,
)
from weaviate.collections.classes.filters import Filter
from weaviate.collections.classes.grpc import (
    HybridFusion,
    FromNested,
    FromReference,
    FromReferenceMultiTarget,
    MetadataQuery,
)
from weaviate.collections.classes.internal import Nested, CrossReference, Reference
from weaviate.collections.classes.tenants import Tenant

__all__ = [
    "Configure",
    "ConfigureUpdate",
    "CrossReference",
    "DataObject",
    "DataType",
    "Filter",
    "HybridFusion",
    "FromNested",
    "FromReference",
    "FromReferenceMultiTarget",
    "MetadataQuery",
    "Multi2VecField",
    "Nested",
    "Property",
    "Reference",
    "ReferenceProperty",
    "ReferencePropertyMultiTarget",
    "Tenant",
    "Tokenization",
    "VectorDistance",
]



import time

import pytest
from requests import ReadTimeout
from werkzeug.wrappers import Request, Response

import weaviate
from mock_tests.conftest import MOCK_SERVER_URL


def test_schema_timeout_error(weaviate_mock):
    """Tests that expected timeout exception is raised."""

    def handler(request: Request):
        time.sleep(1.5)  # cause timeout
        return Response(status=200)

    weaviate_mock.expect_request("/v1/schema/Test").respond_with_handler(handler)
    client = weaviate.Client(MOCK_SERVER_URL, timeout_config=(1, 1))

    with pytest.raises(ReadTimeout):
        client.schema.exists("Test")


def test_schema_unknown_status_code(weaviate_mock):
    """Tests that expected UnexpectedStatusCodeException exception is raised."""

    def handler(request: Request):
        return Response(status=403)

    weaviate_mock.expect_request("/v1/schema/Test").respond_with_handler(handler)
    client = weaviate.Client(MOCK_SERVER_URL)

    with pytest.raises(weaviate.UnexpectedStatusCodeException):
        client.schema.exists("Test")


def test_schema_exists(weaviate_mock):
    """Tests correct behaviour."""

    def handler(request: Request, status: int):
        return Response(status=status)

    weaviate_mock.expect_request("/v1/schema/Exists").respond_with_handler(
        lambda r: handler(r, 200)
    )
    weaviate_mock.expect_request("/v1/schema/DoesNotExists").respond_with_handler(
        lambda r: handler(r, 404)
    )
    client = weaviate.Client(MOCK_SERVER_URL)

    assert client.schema.exists("Exists") is True
    assert client.schema.exists("DoesNotExists") is False

    assert client.schema.exists("exists") is True
    assert client.schema.exists("doesNotExists") is False



from typing import List, Optional, Union

import pytest

from weaviate.collections.classes.grpc import (
    PROPERTIES,
    FromReference,
    FromReferenceMultiTarget,
    MetadataQuery,
)
from weaviate.collections.queries.base import _PropertiesParser


@pytest.mark.parametrize(
    "properties,output",
    [
        (["name"], ["name"]),
        ([FromReference(link_on="name")], [FromReference(link_on="name")]),
        (["name", FromReference(link_on="name")], ["name", FromReference(link_on="name")]),
        (
            [FromReferenceMultiTarget(link_on="name", target_collection="Test"), "name"],
            ["name", FromReferenceMultiTarget(link_on="name", target_collection="Test")],
        ),
        (
            ["__article__properties__name"],
            [FromReference(link_on="article", return_properties=["name"])],
        ),
        (
            ["__article__properties__name", "name", "__article__properties__age"],
            ["name", FromReference(link_on="article", return_properties=["name", "age"])],
        ),
        (
            ["__article__metadata__uuid"],
            [FromReference(link_on="article", return_metadata=MetadataQuery(uuid=True))],
        ),
        (
            ["__article__metadata__uuid", "__article__metadata__vector"],
            [
                FromReference(
                    link_on="article", return_metadata=MetadataQuery(uuid=True, vector=True)
                )
            ],
        ),
        (["__article"], [FromReference(link_on="article")]),
        (
            ["__article", "__article__metadata__uuid"],
            [FromReference(link_on="article", return_metadata=MetadataQuery(uuid=True))],
        ),
        (
            ["__article", "__article__properties__name"],
            [FromReference(link_on="article", return_properties=["name"])],
        ),
        (
            ["__article__metadata__uuid", "__article"],
            [FromReference(link_on="article", return_metadata=MetadataQuery(uuid=True))],
        ),
        (
            ["__article__properties__name", "__article"],
            [FromReference(link_on="article", return_properties=["name"])],
        ),
        (None, None),
        ([None, "name"], ["name"]),
        (["__article", "name"], ["name", FromReference(link_on="article")]),
        ("__article", [FromReference(link_on="article")]),
        ("__article__", [FromReference(link_on="article")]),
    ],
)
def test_properties_parser_success(properties: Optional[PROPERTIES], output: Optional[PROPERTIES]):
    assert output == _PropertiesParser().parse(properties)


ERROR_MESSAGE = (
    lambda s: f"return reference property {s} must be in the format __{{prop_name}} or __{{prop_name}}__{{properties|metadata}}_{{nested_prop_name}} when using string syntax"
)


@pytest.mark.parametrize(
    "wrong",
    [
        ["__article__propertie__uuid"],
        ["__article__propertiess__uuid"],
        ["__article__metadat__uuid"],
        ["__article__metadataa__uuid"],
        "__article__propertie__uuid",
        "__article__propertiess__uuid",
        "__article__metadat__uuid",
        "__article__metadataa__uuid",
    ],
)
def test_properties_parser_error(wrong: Union[str, List[str]]):
    with pytest.raises(ValueError) as e:
        out = _PropertiesParser().parse(wrong)
        print(out)
    assert e.value.args[0] == ERROR_MESSAGE(wrong[0] if len(wrong) == 1 else wrong)



import os
import unittest
from copy import deepcopy
from unittest.mock import patch, Mock

from requests.exceptions import ConnectionError as RequestsConnectionError

from test.util import mock_connection_func, check_error_message, check_startswith_error_message
from weaviate.exceptions import UnexpectedStatusCodeException
from weaviate.schema import Schema
from weaviate.util import _capitalize_first_letter

company_test_schema = {
    "classes": [
        {
            "class": "Company",
            "description": "A business that acts in the market",
            "properties": [
                {
                    "name": "name",
                    "description": "The name under which the company is known",
                    "dataType": ["text"],
                },
                {
                    "name": "legalBody",
                    "description": "The legal body under which the company maintains its business",
                    "dataType": ["text"],
                },
                {
                    "name": "hasEmployee",
                    "description": "The employees of the company",
                    "dataType": ["Employee"],
                },
            ],
        },
        {
            "class": "Employee",
            "description": "An employee of the company",
            "properties": [
                {
                    "name": "name",
                    "description": "The name of the employee",
                    "dataType": ["text"],
                },
                {
                    "name": "job",
                    "description": "the job description of the employee",
                    "dataType": ["text"],
                },
                {
                    "name": "yearsInTheCompany",
                    "description": "The number of years this employee has worked in the company",
                    "dataType": ["int"],
                },
            ],
        },
    ]
}

# A test schema as it was returned from a real weaviate instance
persons_return_test_schema = {
    "classes": [
        {
            "class": "Person",
            "description": "A person such as humans or personality known through culture",
            "properties": [
                {"dataType": ["text"], "description": "The name of this person", "name": "name"}
            ],
        },
        {
            "class": "Group",
            "description": "A set of persons who are associated with each other over some common properties",
            "properties": [
                {
                    "dataType": ["text"],
                    "description": "The name under which this group is known",
                    "name": "name",
                },
                {
                    "dataType": ["Person"],
                    "description": "The persons that are part of this group",
                    "name": "members",
                },
            ],
        },
    ],
}

schema_company_local = {  # NOTE: should be the same as file schema_company.json
    "classes": [
        {
            "class": "Company",
            "description": "A business that acts in the market",
            "properties": [
                {
                    "name": "name",
                    "description": "The name under which the company is known",
                    "dataType": ["text"],
                },
                {
                    "name": "legalBody",
                    "description": "The legal body under which the company maintains its business",
                    "dataType": ["text"],
                },
                {
                    "name": "hasEmployee",
                    "description": "The employees of the company",
                    "dataType": ["Employee"],
                },
            ],
        },
        {
            "class": "Employee",
            "description": "An employee of the company",
            "properties": [
                {"name": "name", "description": "The name of the employee", "dataType": ["text"]},
                {
                    "name": "job",
                    "description": "the job description of the employee",
                    "dataType": ["text"],
                },
                {
                    "name": "yearsInTheCompany",
                    "description": "The number of years this employee has worked in the company",
                    "dataType": ["int"],
                },
            ],
        },
    ]
}


class TestSchema(unittest.TestCase):
    def test_create(self):
        """
        Test the `create` method.
        """

        schema = Schema(Mock())

        # mock function calls
        mock_primitive = Mock()
        mock_complex = Mock()
        schema._create_classes_with_primitives = mock_primitive
        schema._create_complex_properties_from_classes = mock_complex

        schema.create("test/schema/schema_company.json")  # with read from file

        mock_primitive.assert_called_with(schema_company_local["classes"])
        mock_complex.assert_called_with(schema_company_local["classes"])

    def test_create_class(self):
        """
        Test the `create_class` method.
        """

        schema = Schema(Mock())

        # mock function calls
        mock_primitive = Mock()
        mock_complex = Mock()
        schema._create_class_with_primitives = mock_primitive
        schema._create_complex_properties_from_class = mock_complex

        schema.create_class(company_test_schema["classes"][0])

        mock_primitive.assert_called_with(company_test_schema["classes"][0])
        mock_complex.assert_called_with(company_test_schema["classes"][0])

    @patch("weaviate.schema.crud_schema.Schema.get")
    def test_update_config(self, mock_schema):
        """
        Test the `update_config` method.
        """

        # invalid calls
        requests_error_message = "Class schema configuration could not be updated."
        unexpected_error_msg = "Update class schema configuration"

        mock_schema.return_value = {
            "class": "Test",
            "vectorIndexConfig": {"test1": "Test1", "test2": 2},
        }
        mock_conn = mock_connection_func("put", side_effect=RequestsConnectionError("Test!"))
        schema = Schema(mock_conn)
        with self.assertRaises(RequestsConnectionError) as error:
            schema.update_config("Test", {"vectorIndexConfig": {"test2": "Test2"}})
        check_error_message(self, error, requests_error_message)
        mock_conn.put.assert_called_with(
            path="/schema/Test",
            weaviate_object={
                "class": "Test",
                "vectorIndexConfig": {"test1": "Test1", "test2": "Test2"},
            },
        )

        mock_schema.return_value = {
            "class": "Test",
            "vectorIndexConfig": {"test1": "Test1", "test2": 2},
        }
        mock_conn = mock_connection_func("put", status_code=404)
        schema = Schema(mock_conn)
        with self.assertRaises(UnexpectedStatusCodeException) as error:
            schema.update_config("Test", {"vectorIndexConfig": {"test3": True}})
        check_startswith_error_message(self, error, unexpected_error_msg)
        mock_conn.put.assert_called_with(
            path="/schema/Test",
            weaviate_object={
                "class": "Test",
                "vectorIndexConfig": {"test1": "Test1", "test2": 2, "test3": True},
            },
        )

        # valid calls
        mock_schema.return_value = {
            "class": "Test",
            "vectorIndexConfig": {"test1": "Test1", "test2": 2},
        }
        mock_conn = mock_connection_func("put")
        schema = Schema(mock_conn)
        schema.update_config("Test", {})
        mock_conn.put.assert_called_with(
            path="/schema/Test",
            weaviate_object={"class": "Test", "vectorIndexConfig": {"test1": "Test1", "test2": 2}},
        )

        # with uncapitalized class_name
        mock_schema.return_value = {
            "class": "Test",
            "vectorIndexConfig": {"test1": "Test1", "test2": 2},
        }
        mock_conn = mock_connection_func("put")
        schema = Schema(mock_conn)
        schema.update_config("test", {})
        mock_conn.put.assert_called_with(
            path="/schema/Test",
            weaviate_object={"class": "Test", "vectorIndexConfig": {"test1": "Test1", "test2": 2}},
        )

    def test_get(self):
        """
        Test the `get` method.
        """

        # invalid calls
        requests_error_message = "Schema could not be retrieved."
        unexpected_error_msg = "Get schema"
        type_error_msg = lambda dt: f"'class_name' argument must be of type `str`! Given type: {dt}"

        mock_conn = mock_connection_func("get", side_effect=RequestsConnectionError("Test!"))
        schema = Schema(mock_conn)
        with self.assertRaises(RequestsConnectionError) as error:
            schema.get()
        check_error_message(self, error, requests_error_message)

        mock_conn = mock_connection_func("get", status_code=404)
        schema = Schema(mock_conn)
        with self.assertRaises(UnexpectedStatusCodeException) as error:
            schema.get()
        check_startswith_error_message(self, error, unexpected_error_msg)

        connection_mock_file = mock_connection_func(
            "get", status_code=200, return_json={"Test": "OK!"}
        )
        schema = Schema(connection_mock_file)
        with self.assertRaises(TypeError) as error:
            schema.get(1234)
        check_error_message(self, error, type_error_msg(int))

        # valid calls

        self.assertEqual(schema.get(), {"Test": "OK!"})
        connection_mock_file.get.assert_called_with(
            path="/schema",
        )

        self.assertEqual(schema.get("Artist"), {"Test": "OK!"})
        connection_mock_file.get.assert_called_with(path="/schema/Artist")

        # with uncapitalized class_name
        self.assertEqual(schema.get("artist"), {"Test": "OK!"})
        connection_mock_file.get.assert_called_with(path="/schema/Artist")

    def test_contains(self):
        """
        Test the `contains` method.
        """

        # If a schema is present it should return true otherwise false
        # 1. test schema is present:

        schema = Schema(mock_connection_func("get", return_json=persons_return_test_schema))
        self.assertTrue(schema.contains())

        # 2. test no schema is present:

        schema = Schema(mock_connection_func("get", return_json={"classes": []}))
        self.assertFalse(schema.contains())

        # 3. test with 'schema' argument
        ## Test weaviate.schema.contains specific schema.

        schema = Schema(mock_connection_func("get", return_json=persons_return_test_schema))
        self.assertFalse(schema.contains(company_test_schema))
        subset_schema = {
            "classes": [
                {
                    "class": "Person",
                    "description": "",
                    "properties": [{"dataType": ["text"], "description": "", "name": "name"}],
                }
            ]
        }
        self.assertTrue(schema.contains(subset_schema))

        ## Test weaviate.schema.contains schema from file.

        schema = Schema(mock_connection_func("get", return_json=persons_return_test_schema))
        schema_json_file = os.path.join(os.path.dirname(__file__), "schema_company.json")
        self.assertFalse(schema.contains(schema_json_file))

        schema = Schema(mock_connection_func("get", return_json=company_test_schema))
        self.assertTrue(schema.contains(schema_json_file))

    def test_delete_class_input(self):
        """
        Test the 'delete_class` method.
        """

        schema = Schema(Mock())

        # invalid calls
        type_error_message = lambda t: f"Class name was {t} instead of str"
        requests_error_message = "Deletion of class."

        with self.assertRaises(TypeError) as error:
            schema.delete_class(1)
        check_error_message(self, error, type_error_message(int))

        schema = Schema(
            mock_connection_func("delete", side_effect=RequestsConnectionError("Test!"))
        )
        with self.assertRaises(RequestsConnectionError) as error:
            schema.delete_class("uuid")
        check_error_message(self, error, requests_error_message)

        schema = Schema(mock_connection_func("delete", status_code=404))
        with self.assertRaises(UnexpectedStatusCodeException) as error:
            schema.delete_class("uuid")
        check_startswith_error_message(self, error, "Delete class from schema")

        # valid calls
        mock_conn = mock_connection_func("delete", status_code=200)
        schema = Schema(mock_conn)
        schema.delete_class("Test")
        mock_conn.delete.assert_called_with(path="/schema/Test")

        # with uncapitalized class_name
        mock_conn = mock_connection_func("delete", status_code=200)
        schema = Schema(mock_conn)
        schema.delete_class("test")
        mock_conn.delete.assert_called_with(path="/schema/Test")

    def test_delete_everything(self):
        """
        Test the `delete_all` method.
        """

        mock_connection = mock_connection_func("get", return_json=company_test_schema)
        mock_connection = mock_connection_func("delete", connection_mock=mock_connection)
        schema = Schema(mock_connection)

        schema.delete_all()
        self.assertEqual(mock_connection.get.call_count, 1)
        self.assertEqual(mock_connection.delete.call_count, 2)

    def test__create_complex_properties_from_classes(self):
        """
        Test the `_create_complex_properties_from_classes` method.
        """

        schema = Schema(Mock())

        mock_complex = Mock()
        schema._create_complex_properties_from_class = mock_complex

        schema._create_complex_properties_from_classes(list("Test!"))
        self.assertEqual(mock_complex.call_count, 5)

    def test__create_complex_properties_from_class(self):
        """
        Test the `_create_complex_properties_from_class` method.
        """

        # valid calls

        def helper_test(nr_calls=1):
            mock_rest = mock_connection_func("post")
            schema = Schema(mock_rest)
            schema._create_complex_properties_from_class(properties)
            self.assertEqual(mock_rest.post.call_count, nr_calls)
            properties_copy = deepcopy(properties["properties"])
            for prop in properties_copy:
                prop["dataType"] = [_capitalize_first_letter(dt) for dt in prop["dataType"]]
            mock_rest.post.assert_called_with(
                path="/schema/" + _capitalize_first_letter(properties["class"]) + "/properties",
                weaviate_object=properties_copy[0],
            )

        # no `properties` key
        mock_rest = mock_connection_func("post")
        schema = Schema(mock_rest)

        schema._create_complex_properties_from_class({})
        self.assertEqual(mock_rest.run_rest.call_count, 0)

        # no COMPLEX properties
        properties = {"properties": [{"dataType": ["text"]}]}
        schema._create_complex_properties_from_class(properties)
        self.assertEqual(mock_rest.post.call_count, 0)

        properties = {"properties": [{"dataType": ["text"]}, {"dataType": ["string"]}]}
        schema._create_complex_properties_from_class(properties)
        self.assertEqual(mock_rest.post.call_count, 0)

        # COMPLEX properties
        properties = {
            "class": "TestClass",
            "properties": [
                {"dataType": ["Test"], "description": "test description", "name": "test_prop"},
            ],
        }
        mock_rest = mock_connection_func("post")
        schema = Schema(mock_rest)
        schema._create_complex_properties_from_class(properties)
        self.assertEqual(mock_rest.post.call_count, 1)

        properties = {
            "class": "TestClass",
            "properties": [
                {"dataType": ["Test"], "description": "test description", "name": "test_prop"},
            ],
        }
        helper_test()

        properties["properties"][0]["indexInverted"] = True
        helper_test()

        properties["properties"][0]["moduleConfig"] = {"test": "ok!"}
        helper_test()

        properties["properties"].append(properties["properties"][0])  # add another property
        properties["properties"].append(properties["properties"][0])  # add another property
        helper_test(3)

        # with uncapitalized class_name
        properties["class"] = "testClass"
        helper_test(3)

        properties = {
            "class": "testClass",
            "properties": [
                {
                    "dataType": ["test", "myTest"],
                    "description": "test description",
                    "name": "test_prop",
                },
            ],
        }

        # invalid calls
        requests_error_message = "Property may not have been created properly."

        mock_rest = mock_connection_func("post", side_effect=RequestsConnectionError("TEST1"))
        schema = Schema(mock_rest)
        with self.assertRaises(RequestsConnectionError) as error:
            schema._create_complex_properties_from_class(properties)
        check_error_message(self, error, requests_error_message)

        mock_rest = mock_connection_func("post", status_code=404)
        schema = Schema(mock_rest)
        with self.assertRaises(UnexpectedStatusCodeException) as error:
            schema._create_complex_properties_from_class(properties)
        check_startswith_error_message(self, error, "Add properties to classes")

    def test__create_class_with_primitives(self):
        """
        Test the `_create_class_with_primitives` method.
        """

        # valid calls
        def helper_test(test_class, test_class_call):
            mock_rest = mock_connection_func("post")
            schema = Schema(mock_rest)
            schema._create_class_with_primitives(test_class)
            self.assertEqual(mock_rest.post.call_count, 1)
            mock_rest.post.assert_called_with(
                path="/schema",
                weaviate_object=test_class_call,
            )

        test_class = {
            "class": "TestClass",
            "properties": [
                {"dataType": ["int"], "name": "test_prop", "description": "None"},
                {"dataType": ["Test"], "name": "test_prop", "description": "None"},
            ],
        }
        test_class_call = {
            "class": "TestClass",
            "properties": [
                {"dataType": ["int"], "name": "test_prop", "description": "None"},
            ],
        }
        helper_test(test_class, test_class_call)

        test_class["description"] = "description"
        test_class_call["description"] = "description"
        helper_test(test_class, test_class_call)

        test_class["description"] = "description"
        test_class_call["description"] = "description"
        helper_test(test_class, test_class_call)

        test_class["vectorIndexType"] = "vectorIndexType"
        test_class_call["vectorIndexType"] = "vectorIndexType"
        helper_test(test_class, test_class_call)

        test_class["vectorIndexConfig"] = {"vectorIndexConfig": "vectorIndexConfig"}
        test_class_call["vectorIndexConfig"] = {"vectorIndexConfig": "vectorIndexConfig"}
        helper_test(test_class, test_class_call)

        test_class["vectorizer"] = "test_vectorizer"
        test_class_call["vectorizer"] = "test_vectorizer"
        helper_test(test_class, test_class_call)

        test_class["moduleConfig"] = {"moduleConfig": "moduleConfig"}
        test_class_call["moduleConfig"] = {"moduleConfig": "moduleConfig"}
        helper_test(test_class, test_class_call)

        test_class["shardingConfig"] = {"shardingConfig": "shardingConfig"}
        test_class_call["shardingConfig"] = {"shardingConfig": "shardingConfig"}
        helper_test(test_class, test_class_call)

        # multiple properties do not imply multiple `run_rest` calls
        test_class["properties"].append(test_class["properties"][0])  # add another property
        test_class["properties"].append(test_class["properties"][0])  # add another property
        test_class_call["properties"].append(test_class["properties"][0])  # add another property
        test_class_call["properties"].append(test_class["properties"][0])  # add another property
        helper_test(test_class, test_class_call)

        # with uncapitalized class_name
        test_class["class"] = "testClass"
        helper_test(test_class, test_class_call)

        # invalid calls
        requests_error_message = "Class may not have been created properly."

        mock_rest = mock_connection_func("post", side_effect=RequestsConnectionError("TEST1"))
        schema = Schema(mock_rest)
        with self.assertRaises(RequestsConnectionError) as error:
            schema._create_class_with_primitives(test_class)
        check_error_message(self, error, requests_error_message)

        mock_rest = mock_connection_func("post", status_code=404)
        schema = Schema(mock_rest)
        with self.assertRaises(UnexpectedStatusCodeException) as error:
            schema._create_class_with_primitives(test_class)
        check_startswith_error_message(self, error, "Create class")

    def test__create_classes_with_primitives(self):
        """
        Test the `_create_classes_with_primitives` method.
        """

        schema = Schema(Mock())

        mock_primitive = Mock()
        schema._create_class_with_primitives = mock_primitive

        schema._create_classes_with_primitives(list("Test!!"))
        self.assertEqual(mock_primitive.call_count, 6)

    def test__property_is_primitive(self):
        """
        Test the `_property_is_primitive` function.
        """

        from weaviate.schema.crud_schema import _property_is_primitive

        test_types_list = ["NOT Primitive", "Neither this one", "Nor This!"]
        self.assertFalse(_property_is_primitive(test_types_list))
        test_types_list = ["NOT Primitive", "boolean", "text"]
        self.assertFalse(_property_is_primitive(test_types_list))
        test_types_list = ["text"]
        self.assertTrue(_property_is_primitive(test_types_list))
        test_types_list = ["int"]
        self.assertTrue(_property_is_primitive(test_types_list))
        test_types_list = ["number"]
        self.assertTrue(_property_is_primitive(test_types_list))
        test_types_list = ["string"]
        self.assertTrue(_property_is_primitive(test_types_list))
        test_types_list = ["boolean"]
        self.assertTrue(_property_is_primitive(test_types_list))
        test_types_list = ["date"]
        self.assertTrue(_property_is_primitive(test_types_list))
        test_types_list = ["geoCoordinates"]
        self.assertTrue(_property_is_primitive(test_types_list))
        test_types_list = ["blob"]
        self.assertTrue(_property_is_primitive(test_types_list))
        test_types_list = ["phoneNumber"]
        self.assertTrue(_property_is_primitive(test_types_list))
        test_types_list = ["int[]", "number[]", "text[]", "string[]", "boolean[]", "date[]"]
        self.assertTrue(_property_is_primitive(test_types_list))
        test_types_list = ["int()"]
        self.assertFalse(_property_is_primitive(test_types_list))
        test_types_list = ["number()"]
        self.assertFalse(_property_is_primitive(test_types_list))
        test_types_list = ["text()"]
        self.assertFalse(_property_is_primitive(test_types_list))
        test_types_list = ["string()"]
        self.assertFalse(_property_is_primitive(test_types_list))
        test_types_list = ["boolean()"]
        self.assertFalse(_property_is_primitive(test_types_list))
        test_types_list = ["date()"]
        self.assertFalse(_property_is_primitive(test_types_list))
        test_types_list = [
            "string",
            "int",
            "boolean",
            "number",
            "date",
            "text",
            "geoCoordinates",
            "blob",
            "phoneNumber",
            "int[]",
            "number[]",
            "text[]",
            "string[]",
            "boolean[]",
            "date[]",
        ]
        self.assertTrue(_property_is_primitive(test_types_list))

    def test__get_primitive_properties(self):
        """
        Test the `_get_primitive_properties` function.
        """

        from weaviate.schema.crud_schema import _get_primitive_properties

        test_func = _get_primitive_properties

        properties_list = []
        self.assertEqual(test_func(properties_list), properties_list)

        properties_list = [{"dataType": ["text"]}]
        self.assertEqual(test_func(properties_list), properties_list)

        properties_list = [{"dataType": ["text"]}, {"dataType": ["int"]}]
        self.assertEqual(test_func(properties_list), properties_list)

        properties_list = [{"dataType": ["Test1"]}, {"dataType": ["Test2"]}]
        self.assertEqual(test_func(properties_list), [])

        properties_list = [
            {"dataType": ["text"]},
            {"dataType": ["int"]},
            {"dataType": ["Test1"]},
            {"dataType": ["Test2"]},
        ]
        self.assertEqual(
            test_func(properties_list), [{"dataType": ["text"]}, {"dataType": ["int"]}]
        )



"""
DataObject class definition.
"""
import uuid as uuid_lib
import warnings
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union, cast

from requests.exceptions import ConnectionError as RequestsConnectionError

from weaviate.connect import Connection
from weaviate.data.references import Reference
from weaviate.data.replication import ConsistencyLevel
from weaviate.error_msgs import DATA_DEPRECATION_NEW_V14_CLS_NS_W, DATA_DEPRECATION_OLD_V14_CLS_NS_W
from weaviate.exceptions import (
    ObjectAlreadyExistsException,
    UnexpectedStatusCodeException,
)
from weaviate.util import (
    _get_dict_from_object,
    get_vector,
    get_valid_uuid,
    _capitalize_first_letter,
    _check_positive_num,
)
from weaviate.types import UUID


class DataObject:
    """
    DataObject class used to manipulate object to/from Weaviate.

    Attributes
    ----------
    reference : weaviate.data.references.Reference
        A Reference object to create objects cross-references.
    """

    def __init__(self, connection: Connection):
        """
        Initialize a DataObject class instance.

        Parameters
        ----------
        connection : weaviate.connect.Connection
            Connection object to an active and running Weaviate instance.
        """

        self._connection = connection
        self.reference = Reference(self._connection)

    def create(
        self,
        data_object: Union[dict, str],
        class_name: str,
        uuid: Union[str, uuid_lib.UUID, None] = None,
        vector: Optional[Sequence] = None,
        consistency_level: Optional[ConsistencyLevel] = None,
        tenant: Optional[str] = None,
    ) -> str:
        """
        Takes a dict describing the object and adds it to Weaviate.

        Parameters
        ----------
        data_object : dict or str
            Object to be added.
            If type is str it should be either a URL or a file.
        class_name : str
            Class name associated with the object given.
        uuid : str, uuid.UUID or None, optional
            Object will be created under this uuid if it is provided.
            Otherwise, Weaviate will generate a uuid for this object,
            by default None.
        vector: Sequence or None, optional
            Embedding for the object.
            Can be used when:
             - a class does not have a vectorization module.
             - The given vector was generated using the _identical_ vectorization module that is configured for the
             class. In this case this vector takes precedence.

            Supported types are `list`, 'numpy.ndarray`, `torch.Tensor` and `tf.Tensor`,
            by default None.
        consistency_level : Optional[ConsistencyLevel], optional
            Can be one of 'ALL', 'ONE', or 'QUORUM'. Determines how many replicas must acknowledge
        tenant: Optional[str], optional
            The name of the tenant for which this operation is being performed.

        Examples
        --------
        Schema contains a class Author with only 'name' and 'age' primitive property.

        >>> client.data_object.create(
        ...     data_object = {'name': 'Neil Gaiman', 'age': 60},
        ...     class_name = 'Author',
        ... )
        '46091506-e3a0-41a4-9597-10e3064d8e2d'
        >>> client.data_object.create(
        ...     data_object = {'name': 'Andrzej Sapkowski', 'age': 72},
        ...     class_name = 'Author',
        ...     uuid = 'e067f671-1202-42c6-848b-ff4d1eb804ab'
        ... )
        'e067f671-1202-42c6-848b-ff4d1eb804ab'

        Returns
        -------
        str
            Returns the UUID of the created object if successful.

        Raises
        ------
        TypeError
            If argument is of wrong type.
        ValueError
            If argument contains an invalid value.
        weaviate.ObjectAlreadyExistsException
            If an object with the given uuid already exists within Weaviate.
        weaviate.UnexpectedStatusCodeException
            If creating the object in Weaviate failed for a different reason,
            more information is given in the exception.
        requests.ConnectionError
            If the network connection to Weaviate fails.
        """

        if not isinstance(class_name, str):
            raise TypeError(f"Expected class_name of type str but was: {type(class_name)}")
        loaded_data_object = _get_dict_from_object(data_object)

        weaviate_obj = {
            "class": _capitalize_first_letter(class_name),
            "properties": loaded_data_object,
        }
        if uuid is not None:
            weaviate_obj["id"] = get_valid_uuid(uuid)

        if vector is not None:
            weaviate_obj["vector"] = get_vector(vector)

        path = "/objects"
        params = {}
        if consistency_level is not None:
            params["consistency_level"] = ConsistencyLevel(consistency_level).value
        if tenant is not None:
            weaviate_obj["tenant"] = tenant
        try:
            response = self._connection.post(path=path, weaviate_object=weaviate_obj, params=params)
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError("Object was not added to Weaviate.") from conn_err
        if response.status_code == 200:
            return str(response.json()["id"])

        object_does_already_exist = False
        try:
            if "already exists" in response.json()["error"][0]["message"]:
                object_does_already_exist = True
        except KeyError:
            pass
        if object_does_already_exist:
            raise ObjectAlreadyExistsException(str(uuid))
        raise UnexpectedStatusCodeException("Creating object", response)

    def update(
        self,
        data_object: Union[dict, str],
        class_name: str,
        uuid: Union[str, uuid_lib.UUID],
        vector: Optional[Sequence] = None,
        consistency_level: Optional[ConsistencyLevel] = None,
        tenant: Optional[str] = None,
    ) -> None:
        """
        Update an already existing object in Weaviate with the given data object.
        Overwrites only the specified fields, the unspecified ones remain unchanged.

        Parameters
        ----------
        data_object : dict or str
            The object states the fields that should be updated.
            Fields not specified in the 'data_object' remain unchanged.
            Fields that are None will not be changed.
            If type is str it should be either an URL or a file.
        class_name : str
            The class name of the object.
        uuid : str or uuid.UUID
            The ID of the object that should be changed.
        vector: Sequence or None, optional
            Embedding for the object.
            Can be used when:
             - a class does not have a vectorization module.
             - The given vector was generated using the _identical_ vectorization module that is configured for the
             class. In this case this vector takes precedence.

            Supported types are `list`, 'numpy.ndarray`, `torch.Tensor` and `tf.Tensor`,
            by default None.
        consistency_level : Optional[ConsistencyLevel], optional
            Can be one of 'ALL', 'ONE', or 'QUORUM'. Determines how many replicas must acknowledge
        tenant: Optional[str], optional
            The name of the tenant for which this operation is being performed.

        Examples
        --------
        >>> author_id = client.data_object.create(
        ...     data_object = {'name': 'Philip Pullman', 'age': 64},
        ...     class_name = 'Author'
        ... )
        >>> client.data_object.get(author_id)
        {
            "additional": {},
            "class": "Author",
            "creationTimeUnix": 1617111215172,
            "id": "bec2bca7-264f-452a-a5bb-427eb4add068",
            "lastUpdateTimeUnix": 1617111215172,
            "properties": {
                "age": 64,
                "name": "Philip Pullman"
            },
            "vectorWeights": null
        }
        >>> client.data_object.update(
        ...     data_object = {'age': 74},
        ...     class_name = 'Author',
        ...     uuid = author_id
        ... )
        >>> client.data_object.get(author_id)
        {
            "additional": {},
            "class": "Author",
            "creationTimeUnix": 1617111215172,
            "id": "bec2bca7-264f-452a-a5bb-427eb4add068",
            "lastUpdateTimeUnix": 1617111215172,
            "properties": {
                "age": 74,
                "name": "Philip Pullman"
            },
            "vectorWeights": null
        }

        Raises
        ------
        TypeError
            If argument is of wrong type.
        ValueError
            If argument contains an invalid value.
        requests.ConnectionError
            If the network connection to Weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If Weaviate reports a none successful status.
        """
        params = {}
        if consistency_level is not None:
            params["consistency_level"] = ConsistencyLevel(consistency_level).value
        weaviate_obj, path = self._create_object_for_update(data_object, class_name, uuid, vector)
        if tenant is not None:
            weaviate_obj["tenant"] = tenant

        try:
            response = self._connection.patch(
                path=path,
                weaviate_object=weaviate_obj,
                params=params,
            )
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError("Object was not updated.") from conn_err
        if response.status_code == 204:
            # Successful merge
            return
        raise UnexpectedStatusCodeException("Update of the object not successful", response)

    def replace(
        self,
        data_object: Union[dict, str],
        class_name: str,
        uuid: Union[str, uuid_lib.UUID],
        vector: Optional[Sequence] = None,
        consistency_level: Optional[ConsistencyLevel] = None,
        tenant: Optional[str] = None,
    ) -> None:
        """
        Replace an already existing object with the given data object.
        This method replaces the whole object.

        Parameters
        ----------
        data_object : dict or str
            Describes the new values. It may be an URL or path to a json
            or a python dict describing the new values.
        class_name : str
            Name of the class of the object that should be updated.
        uuid : str or uuid.UUID
            The UUID of the object that should be changed.
        vector: Sequence or None, optional
            Embedding for the object.
            Can be used when:
             - a class does not have a vectorization module.
             - The given vector was generated using the _identical_ vectorization module that is configured for the
             class. In this case this vector takes precedence.

            Supported types are `list`, 'numpy.ndarray`, `torch.Tensor` and `tf.Tensor`,
            by default None.
        consistency_level : Optional[ConsistencyLevel], optional
            Can be one of 'ALL', 'ONE', or 'QUORUM'. Determines how many replicas must acknowledge
        tenant: Optional[str], optional
            The name of the tenant for which this operation is being performed.

        Examples
        --------
        >>> author_id = client.data_object.create(
        ...     data_object = {'name': 'H. Lovecraft', 'age': 46},
        ...     class_name = 'Author'
        ... )
        >>> client.data_object.get(author_id)
        {
            "additional": {},
            "class": "Author",
            "creationTimeUnix": 1617112817487,
            "id": "d842a0f4-ad8c-40eb-80b4-bfefc7b1b530",
            "lastUpdateTimeUnix": 1617112817487,
            "properties": {
                "age": 46,
                "name": "H. Lovecraft"
            },
            "vectorWeights": null
        }
        >>> client.data_object.replace(
        ...     data_object = {'name': 'H.P. Lovecraft'},
        ...     class_name = 'Author',
        ...     uuid = author_id
        ... )
        >>> client.data_object.get(author_id)
        {
            "additional": {},
            "class": "Author",
            "id": "d842a0f4-ad8c-40eb-80b4-bfefc7b1b530",
            "lastUpdateTimeUnix": 1617112838668,
            "properties": {
                "name": "H.P. Lovecraft"
            },
            "vectorWeights": null
        }

        Raises
        ------
        TypeError
            If argument is of wrong type.
        ValueError
            If argument contains an invalid value.
        requests.ConnectionError
            If the network connection to Weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If Weaviate reports a none OK status.
        """
        params = {}
        if consistency_level is not None:
            params["consistency_level"] = ConsistencyLevel(consistency_level).value
        weaviate_obj, path = self._create_object_for_update(data_object, class_name, uuid, vector)
        if tenant is not None:
            weaviate_obj["tenant"] = tenant
        try:
            response = self._connection.put(path=path, weaviate_object=weaviate_obj, params=params)
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError("Object was not replaced.") from conn_err
        if response.status_code == 200:
            # Successful update
            return
        raise UnexpectedStatusCodeException("Replace object", response)

    def _create_object_for_update(
        self,
        data_object: Union[dict, str],
        class_name: str,
        uuid: Union[str, uuid_lib.UUID],
        vector: Optional[Sequence] = None,
    ) -> Tuple[Dict[str, Any], str]:
        if not isinstance(class_name, str):
            raise TypeError("Class must be type str")

        uuid = get_valid_uuid(uuid)

        object_dict = _get_dict_from_object(data_object)

        weaviate_obj = {
            "id": uuid,
            "properties": object_dict,
            "class": _capitalize_first_letter(class_name),
        }

        if vector is not None:
            weaviate_obj["vector"] = get_vector(vector)

        is_server_version_14 = self._connection.server_version >= "1.14"

        if is_server_version_14:
            path = f"/objects/{_capitalize_first_letter(class_name)}/{uuid}"
        else:
            path = f"/objects/{uuid}"
        return weaviate_obj, path

    def get_by_id(
        self,
        uuid: Union[str, uuid_lib.UUID],
        additional_properties: Optional[List[str]] = None,
        with_vector: bool = False,
        class_name: Optional[str] = None,
        node_name: Optional[str] = None,
        consistency_level: Optional[ConsistencyLevel] = None,
        tenant: Optional[str] = None,
    ) -> Optional[dict]:
        """
        Get an object as dict.

        Parameters
        ----------
        uuid : str or uuid.UUID
            The identifier of the object that should be retrieved.
        additional_properties : list of str, optional
            List of additional properties that should be included in the request,
            by default None
        with_vector: bool
            If True the `vector` property will be returned too,
            by default False.
        class_name : Optional[str], optional
            The class name of the object with UUID `uuid`. Introduced in Weaviate version v1.14.0.
            STRONGLY recommended to set it with Weaviate >= 1.14.0. It will be required in future
            versions of Weaviate Server and Clients. Use None value ONLY for Weaviate < v1.14.0,
            by default None
        tenant: str, optional
            The name of the tenant for which this operation is being performed.

        Examples
        --------
        >>> client.data_object.get_by_id(
        ...     uuid="d842a0f4-ad8c-40eb-80b4-bfefc7b1b530",
        ...     class_name='Author', # ONLY with Weaviate >= 1.14.0
        ... )
        {
            "additional": {},
            "class": "Author",
            "creationTimeUnix": 1617112817487,
            "id": "d842a0f4-ad8c-40eb-80b4-bfefc7b1b530",
            "lastUpdateTimeUnix": 1617112817487,
            "properties": {
                "age": 46,
                "name": "H.P. Lovecraft"
            },
            "vectorWeights": null
        }

        Returns
        -------
        dict or None
            dict in case the object exists.
            None in case the object does not exist.

        Raises
        ------
        TypeError
            If argument is of wrong type.
        ValueError
            If argument contains an invalid value.
        requests.ConnectionError
            If the network connection to Weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If Weaviate reports a none OK status.
        """

        return self.get(
            uuid=uuid,
            additional_properties=additional_properties,
            with_vector=with_vector,
            class_name=class_name,
            node_name=node_name,
            consistency_level=consistency_level,
            tenant=tenant,
        )

    def get(
        self,
        uuid: Union[str, uuid_lib.UUID, None] = None,
        additional_properties: Optional[List[str]] = None,
        with_vector: bool = False,
        class_name: Optional[str] = None,
        node_name: Optional[str] = None,
        consistency_level: Optional[ConsistencyLevel] = None,
        limit: Optional[int] = None,
        after: Optional[UUID] = None,
        offset: Optional[int] = None,
        sort: Optional[Dict[str, Union[str, bool, List[bool], List[str]]]] = None,
        tenant: Optional[str] = None,
    ) -> Optional[Dict[str, Any]]:
        """
        Gets objects from Weaviate, the maximum number of objects returned is 100.
        If 'uuid' is None, all objects are returned. If 'uuid' is specified the result is the same
        as for `get_by_uuid` method.

        Parameters
        ----------
        uuid : str, uuid.UUID or None, optional
            The identifier of the object that should be retrieved.
        additional_properties : list of str, optional
            list of additional properties that should be included in the request,
            by default None
        with_vector : bool
            If True the `vector` property will be returned too,
            by default False
        class_name: Optional[str], optional
            The class name of the object with UUID `uuid`. Introduced in Weaviate version v1.14.0.
            STRONGLY recommended to set it with Weaviate >= 1.14.0. It will be required in future
            versions of Weaviate Server and Clients. Use None value ONLY for Weaviate < v1.14.0,
            by default None
        consistency_level : Optional[ConsistencyLevel], optional
            Can be one of 'ALL', 'ONE', or 'QUORUM'. Determines how many replicas must acknowledge
            a request before it is considered successful. Mutually exclusive with node_name param.
        node_name : Optional[str], optional
            The name of the target node which should fulfill the request. Mutually exclusive with
            consistency_level param.
        limit: Optional[int], optional
            The maximum number of data objects to return.
            by default None, which uses the Weaviate default of 100 entries
        after: Optional[UUID], optional
           Can be used to extract all elements by giving the last ID from the previous "page". Requires limit to be set
           but cannot be combined with any other filters or search. Part of the Cursor API.
        offset: Optional[int], optional
            The offset of objects returned, i.e. the starting index of the returned objects. Should be
            used in conjunction with the 'limit' parameter.
        sort: Optional[Dict]
            A dictionary for sorting objects.
            sort['properties']: str, List[str]
                By which properties the returned objects should be sorted. When more than one property is given, the objects are sorted in order of the list.
                The order of the sorting can be given by using 'sort['order_asc']'.
            sort['order_asc']: bool, List[bool]
                The order the properties given in 'sort['properties']' should be returned in. When a single boolean is used, all properties are sorted in the same order.
                If a list is used, it needs to have the same length as 'sort'. Each properties order is then decided individually.
                If 'sort['order_asc']' is True, the properties are sorted in ascending order. If it is False, they are sorted in descending order.
                if 'sort['order_asc']' is not given, all properties are sorted in ascending order.
        tenant: Optional[str], optional
            The name of the tenant for which this operation is being performed.

        Returns
        -------
        list of dicts
            A list of all objects. If no objects where found the list is empty.

        Raises
        ------
        TypeError
            If argument is of wrong type.
        ValueError
            If argument contains an invalid value.
        requests.ConnectionError
            If the network connection to Weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If Weaviate reports a none OK status.
        """
        is_server_version_14 = self._connection.server_version >= "1.14"

        if class_name is None and is_server_version_14 and uuid is not None:
            warnings.warn(
                message=DATA_DEPRECATION_NEW_V14_CLS_NS_W,
                category=DeprecationWarning,
                stacklevel=1,
            )
        if class_name is not None and uuid is not None:
            if not is_server_version_14:
                warnings.warn(
                    message=DATA_DEPRECATION_OLD_V14_CLS_NS_W,
                    category=DeprecationWarning,
                    stacklevel=1,
                )
            if not isinstance(class_name, str):
                raise TypeError(f"'class_name' must be of type str. Given type: {type(class_name)}")

        params = _get_params(additional_properties, with_vector)

        if class_name and is_server_version_14:
            if uuid is not None:
                path = f"/objects/{_capitalize_first_letter(class_name)}"
            else:
                path = "/objects"
                params["class"] = _capitalize_first_letter(class_name)
        else:
            path = "/objects"

        if uuid is not None:
            path += "/" + get_valid_uuid(uuid)

        if consistency_level is not None:
            params["consistency_level"] = ConsistencyLevel(consistency_level).value

        if tenant is not None:
            params["tenant"] = tenant

        if node_name is not None:
            params["node_name"] = node_name

        if limit is not None:
            _check_positive_num(limit, "limit", int, include_zero=False)
            params["limit"] = limit

        if after is not None:
            params["after"] = get_valid_uuid(after)

        if offset is not None:
            _check_positive_num(offset, "offset", int, include_zero=True)
            params["offset"] = offset

        if sort is not None:
            if "properties" not in sort:
                raise ValueError("The sort clause is missing the required field: 'properties'.")
            if "order_asc" not in sort:
                sort["order_asc"] = True
            if not isinstance(sort, Dict):
                raise TypeError(f"'sort' must be of type dict. Given type: {type(sort)}.")
            if isinstance(sort["properties"], str):
                sort["properties"] = [sort["properties"]]
            elif not isinstance(sort["properties"], list) or not all(
                isinstance(x, str) for x in sort["properties"]
            ):
                raise TypeError(
                    f"'sort['properties']' must be of type str or list[str]. Given type: {type(sort['properties'])}."
                )
            if len(sort["properties"]) == 0:
                raise ValueError("'sort['properties']' cannot be an empty list.")

            if isinstance(sort["order_asc"], bool):
                sort["order_asc"] = [sort["order_asc"]] * len(sort["properties"])
            elif not isinstance(sort["order_asc"], list) or not all(
                isinstance(x, bool) for x in sort["order_asc"]
            ):
                raise TypeError(
                    f"'sort['order_asc']' must be of type boolean or list[bool]. Given type: {type(sort['order_asc'])}."
                )
            if len(sort["properties"]) != len(sort["order_asc"]):  # type: ignore
                raise ValueError(
                    f"'sort['order_asc']' must be the same length as 'sort['properties']' or a boolean (not in a list). Current length is sort['properties']:{len(sort['properties'])} and sort['order_asc']:{len(sort['order_asc'])}."  # type: ignore
                )
            if len(sort["order_asc"]) == 0:  # type: ignore
                raise ValueError("'sort['order_asc']' cannot be an empty list.")

            params["sort"] = ",".join(sort["properties"])  # type: ignore
            order = ["asc" if x else "desc" for x in sort["order_asc"]]  # type: ignore
            params["order"] = ",".join(order)

        try:
            response = self._connection.get(
                path=path,
                params=params,
            )
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError("Could not get object/s.") from conn_err
        if response.status_code == 200:
            return cast(Dict[str, Any], response.json())
        if response.status_code == 404:
            return None
        raise UnexpectedStatusCodeException("Get object/s", response)

    def delete(
        self,
        uuid: Union[str, uuid_lib.UUID],
        class_name: Optional[str] = None,
        consistency_level: Optional[ConsistencyLevel] = None,
        tenant: Optional[str] = None,
    ) -> None:
        """
        Delete an existing object from Weaviate.

        Parameters
        ----------
        uuid : str or uuid.UUID
            The ID of the object that should be deleted.
        class_name : Optional[str], optional
            The class name of the object with UUID `uuid`. Introduced in Weaviate version v1.14.0.
            STRONGLY recommended to set it with Weaviate >= 1.14.0. It will be required in future
            versions of Weaviate Server and Clients. Use None value ONLY for Weaviate < v1.14.0,
            by default None
        consistency_level : Optional[ConsistencyLevel], optional
            Can be one of 'ALL', 'ONE', or 'QUORUM'. Determines how many replicas must acknowledge
        tenant: str, optional
            The name of the tenant for which this operation is being performed.

        Examples
        --------
        >>> client.data_object.get(
        ...     uuid="d842a0f4-ad8c-40eb-80b4-bfefc7b1b530",
        ...     class_name='Author', # ONLY with Weaviate >= 1.14.0
        ... )
        {
            "additional": {},
            "class": "Author",
            "creationTimeUnix": 1617112817487,
            "id": "d842a0f4-ad8c-40eb-80b4-bfefc7b1b530",
            "lastUpdateTimeUnix": 1617112817487,
            "properties": {
                "age": 46,
                "name": "H.P. Lovecraft"
            },
            "vectorWeights": null
        }
        >>> client.data_object.delete(
        ...     uuid="d842a0f4-ad8c-40eb-80b4-bfefc7b1b530",
        ...     class_name='Author', # ONLY with Weaviate >= 1.14.0
        ... )
        >>> client.data_object.get(
        ...     uuid="d842a0f4-ad8c-40eb-80b4-bfefc7b1b530",
        ...     class_name='Author', # ONLY with Weaviate >= 1.14.0
        ... )
        None

        Raises
        ------
        requests.ConnectionError
            If the network connection to Weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If Weaviate reports a none OK status.
        TypeError
            If parameter has the wrong type.
        ValueError
            If uuid is not properly formed.
        """

        uuid = get_valid_uuid(uuid)

        is_server_version_14 = self._connection.server_version >= "1.14"

        if class_name is None and is_server_version_14:
            warnings.warn(
                message=DATA_DEPRECATION_NEW_V14_CLS_NS_W,
                category=DeprecationWarning,
                stacklevel=1,
            )
        if class_name is not None:
            if not is_server_version_14:
                warnings.warn(
                    message=DATA_DEPRECATION_OLD_V14_CLS_NS_W,
                    category=DeprecationWarning,
                    stacklevel=1,
                )
            if not isinstance(class_name, str):
                raise TypeError(f"'class_name' must be of type str. Given type: {type(class_name)}")

        if class_name and is_server_version_14:
            path = f"/objects/{_capitalize_first_letter(class_name)}/{uuid}"
        else:
            path = f"/objects/{uuid}"

        params = {}
        if consistency_level is not None:
            params = {"consistency_level": ConsistencyLevel(consistency_level).value}
        if tenant is not None:
            params["tenant"] = tenant
        try:
            response = self._connection.delete(
                path=path,
                params=params,
            )
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError("Object could not be deleted.") from conn_err
        if response.status_code == 204:
            # Successfully deleted
            return
        raise UnexpectedStatusCodeException("Delete object", response)

    def exists(
        self,
        uuid: Union[str, uuid_lib.UUID],
        class_name: Optional[str] = None,
        consistency_level: Optional[ConsistencyLevel] = None,
        tenant: Optional[str] = None,
    ) -> bool:
        """
        Check if the object exist in Weaviate.

        Parameters
        ----------
        uuid : str or uuid.UUID
            The UUID of the object that may or may not exist within Weaviate.
        class_name : Optional[str], optional
            The class name of the object with UUID `uuid`. Introduced in Weaviate version 1.14.0.
            STRONGLY recommended to set it with Weaviate >= 1.14.0. It will be required in future
            versions of Weaviate Server and Clients. Use None value ONLY for Weaviate < 1.14.0,
            by default None
        consistency_level : Optional[ConsistencyLevel], optional
            Can be one of 'ALL', 'ONE', or 'QUORUM'. Determines how many replicas must acknowledge
        tenant: Optional[str], optional
            The name of the tenant for which this operation is being performed.

        Examples
        --------
        >>> client.data_object.exists(
        ...     uuid='e067f671-1202-42c6-848b-ff4d1eb804ab',
        ...     class_name='Author',  # ONLY with Weaviate >= 1.14.0
        ... )
        False
        >>> client.data_object.create(
        ...     data_object = {'name': 'Andrzej Sapkowski', 'age': 72},
        ...     class_name = 'Author',
        ...     uuid = 'e067f671-1202-42c6-848b-ff4d1eb804ab'
        ... )
        >>> client.data_object.exists(
        ...     uuid='e067f671-1202-42c6-848b-ff4d1eb804ab',
        ...     class_name='Author', # ONLY with Weaviate >= 1.14.0
        ... )
        True

        Returns
        -------
        bool
            True if object exists, False otherwise.

        Raises
        ------
        requests.ConnectionError
            If the network connection to Weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If Weaviate reports a none OK status.
        TypeError
            If parameter has the wrong type.
        ValueError
            If uuid is not properly formed.
        """

        is_server_version_14 = self._connection.server_version >= "1.14"

        if class_name is None and is_server_version_14:
            warnings.warn(
                message=DATA_DEPRECATION_NEW_V14_CLS_NS_W,
                category=DeprecationWarning,
                stacklevel=1,
            )
        if class_name is not None:
            if not is_server_version_14:
                warnings.warn(
                    message=DATA_DEPRECATION_OLD_V14_CLS_NS_W,
                    category=DeprecationWarning,
                    stacklevel=1,
                )
            if not isinstance(class_name, str):
                raise TypeError(f"'class_name' must be of type str. Given type: {type(class_name)}")

        if class_name and is_server_version_14:
            path = f"/objects/{_capitalize_first_letter(class_name)}/{get_valid_uuid(uuid)}"
        else:
            path = f"/objects/{get_valid_uuid(uuid)}"
        params = {}
        if consistency_level is not None:
            params = {"consistency_level": ConsistencyLevel(consistency_level).value}
        if tenant is not None:
            params["tenant"] = tenant

        try:
            response = self._connection.head(
                path=path,
                params=params,
            )
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError("Could not check if object exist.") from conn_err

        if response.status_code == 204:
            return True
        if response.status_code == 404:
            return False
        raise UnexpectedStatusCodeException("Object exists", response)

    def validate(
        self,
        data_object: Union[dict, str],
        class_name: str,
        uuid: Union[str, uuid_lib.UUID, None] = None,
        vector: Optional[Sequence] = None,
    ) -> dict:
        """
        Validate an object against Weaviate.

        Parameters
        ----------
        data_object : dict or str
            Object to be validated.
            If type is str it should be either an URL or a file.
        class_name : str
            Name of the class of the object that should be validated.
        uuid : str, uuid.UUID or None, optional
            The UUID of the object that should be validated against Weaviate.
            by default None.
        vector: Sequence or None, optional
            The embedding of the object that should be validated.
            Can be used when:
             - a class does not have a vectorization module.
             - The given vector was generated using the _identical_ vectorization module that is configured for the
             class. In this case this vector takes precedence.

            Supported types are `list`, 'numpy.ndarray`, `torch.Tensor` and `tf.Tensor`,
            by default None.

        Examples
        --------
        Assume we have a Author class only 'name' property, NO 'age'.

        >>> client1.data_object.validate(
        ...     data_object = {'name': 'H. Lovecraft'},
        ...     class_name = 'Author'
        ... )
        {'error': None, 'valid': True}
        >>> client1.data_object.validate(
        ...     data_object = {'name': 'H. Lovecraft', 'age': 46},
        ...     class_name = 'Author'
        ... )
        {
            "error": [
                {
                "message": "invalid object: no such prop with name 'age' found in class 'Author'
                    in the schema. Check your schema files for which properties in this class are
                    available"
                }
            ],
            "valid": false
        }

        Returns
        -------
        dict
            Validation result. E.g. {"valid": bool, "error": None or list}

        Raises
        ------
        TypeError
            If argument is of wrong type.
        ValueError
            If argument contains an invalid value.
        weaviate.UnexpectedStatusCodeException
            If validating the object against Weaviate failed with a different reason.
        requests.ConnectionError
            If the network connection to Weaviate fails.
        """

        loaded_data_object = _get_dict_from_object(data_object)
        if not isinstance(class_name, str):
            raise TypeError(f"Expected class_name of type `str` but was: {type(class_name)}")

        weaviate_obj = {
            "class": _capitalize_first_letter(class_name),
            "properties": loaded_data_object,
        }

        if uuid is not None:
            weaviate_obj["id"] = get_valid_uuid(uuid)

        if vector is not None:
            weaviate_obj["vector"] = get_vector(vector)

        path = "/objects/validate"
        try:
            response = self._connection.post(path=path, weaviate_object=weaviate_obj)
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError(
                "Object was not validated against Weaviate."
            ) from conn_err

        result: dict = {"error": None}

        if response.status_code == 200:
            result["valid"] = True
            return result
        if response.status_code == 422:
            result["valid"] = False
            result["error"] = response.json()["error"]
            return result
        raise UnexpectedStatusCodeException("Validate object", response)


def _get_params(additional_properties: Optional[List[str]], with_vector: bool) -> dict:
    """
    Get underscore properties in the format accepted by Weaviate.

    Parameters
    ----------
    additional_properties : list of str or None
        A list of additional properties or None.
    with_vector: bool
        If True the `vector` property will be returned too.

    Returns
    -------
    dict
        A dictionary including Weaviate-accepted additional properties
        and/or `vector` property.

    Raises
    ------
    TypeError
        If 'additional_properties' is not of type list.
    """

    params = {}
    if additional_properties:
        if not isinstance(additional_properties, list):
            raise TypeError(
                "Additional properties must be of type list "
                f"but are {type(additional_properties)}"
            )
        params["include"] = ",".join(additional_properties)

    if with_vector:
        if "include" in params:
            params["include"] = params["include"] + ",vector"
        else:
            params["include"] = "vector"
    return params



from typing import Any, Dict, Mapping, Optional, Type, get_origin
from typing_extensions import TypeVar, is_typeddict

from pydantic import BaseModel, ConfigDict

from weaviate.exceptions import InvalidDataModelException

Properties = TypeVar("Properties", bound=Mapping[str, Any], default=Dict[str, Any])
"""`Properties` is used wherever a single generic type is needed"""

TProperties = TypeVar("TProperties", bound=Mapping[str, Any], default=Dict[str, Any])
"""`TProperties` is used alongside `Properties` wherever there are two generic types needed

E.g., in `_DataCollection`, `Properties` is used when defining the generic of the class while
`TProperties` is used when defining the generic to be supplied in `.with_data_model` to create
a new instance of `_DataCollection` with a different `Properties` type.

To be clear: `_DataCollection[Properties]().with_data_model(TProperties) -> _DataCollection[TProperties]()`
"""

P = TypeVar("P")
"""`P` is a completely general type that is used wherever generic properties objects are defined that can be used
within the non-ORM and ORM APIs interchangeably"""

T = TypeVar("T")
"""`T` is a completely general type that is used in any kind of generic"""


def _check_data_model(data_model: Optional[Type[Properties]]) -> None:
    if (
        data_model is not None
        and get_origin(data_model) is not dict
        and not is_typeddict(data_model)
    ):
        raise InvalidDataModelException()


class _WeaviateInput(BaseModel):
    model_config = ConfigDict(extra="forbid")



import unittest
from unittest.mock import Mock

from requests.exceptions import ConnectionError as RequestsConnectionError

from test.util import mock_connection_func, check_error_message, check_startswith_error_message
from weaviate.contextionary import Contextionary
from weaviate.exceptions import UnexpectedStatusCodeException


class TestText2VecContextionary(unittest.TestCase):
    def test_extend(self):
        """
        Test `extend` method.
        """

        contextionary = Contextionary(Mock())

        some_concept = {
            "concept": "lsd",
            "definition": "In probability and statistics, the logarithmic series distribution is a discrete probability distribution derived from the Maclaurin series expansion",
        }

        # error messages
        concept_type_error_message = "Concept must be string"
        definition_type_error_message = "Definition must be string"
        weight_type_error_message = "Weight must be float"
        weight_value_error_message = "Weight out of limits 0.0 <= weight <= 1.0"
        requests_error_message = "text2vec-contextionary could not be extended."
        unexpected_error_message = "Extend text2vec-contextionary"

        ## test exceptions
        with self.assertRaises(TypeError) as error:
            contextionary.extend(concept=None, definition=some_concept["definition"], weight=1.0)
        check_error_message(self, error, concept_type_error_message)

        with self.assertRaises(TypeError) as error:
            contextionary.extend(concept=some_concept["concept"], definition=None, weight=1.0)
        check_error_message(self, error, definition_type_error_message)

        with self.assertRaises(TypeError) as error:
            contextionary.extend(**some_concept, weight=None)
        check_error_message(self, error, weight_type_error_message)

        with self.assertRaises(ValueError) as error:
            contextionary.extend(**some_concept, weight=1.1)
        check_error_message(self, error, weight_value_error_message)

        with self.assertRaises(ValueError) as error:
            contextionary.extend(**some_concept, weight=-1.0)
        check_error_message(self, error, weight_value_error_message)

        ## test UnexpectedStatusCodeException
        contextionary = Contextionary(mock_connection_func("post", status_code=404))
        with self.assertRaises(UnexpectedStatusCodeException) as error:
            contextionary.extend(**some_concept)
        check_startswith_error_message(self, error, unexpected_error_message)

        ## test requests error
        contextionary = Contextionary(
            mock_connection_func("post", side_effect=RequestsConnectionError("Test!"))
        )
        with self.assertRaises(RequestsConnectionError) as error:
            contextionary.extend(**some_concept)
        check_error_message(self, error, requests_error_message)

        ## test valid call without specifying 'weight'
        some_concept["weight"] = 1.0
        connection_mock = mock_connection_func("post", status_code=200)
        contextionary = Contextionary(connection_mock)
        contextionary.extend(**some_concept)
        connection_mock.post.assert_called_with(
            path="/modules/text2vec-contextionary/extensions",
            weaviate_object=some_concept,
        )

        ## test valid call with specifying 'weight as error'
        connection_mock = mock_connection_func("post", status_code=200)
        contextionary = Contextionary(connection_mock)
        # add weight to 'some_concept'
        some_concept["weight"] = 0.1234
        contextionary.extend(**some_concept)
        connection_mock.post.assert_called_with(
            path="/modules/text2vec-contextionary/extensions",
            weaviate_object=some_concept,
        )

    def test_get_concept_vector(self):
        """
        Test `get_concept_vector` method.
        """

        # test valid call
        connection_mock = mock_connection_func("get", return_json={"A": "B"})
        contextionary = Contextionary(connection_mock)
        self.assertEqual("B", contextionary.get_concept_vector("sauce")["A"])
        connection_mock.get.assert_called_with(
            path="/modules/text2vec-contextionary/concepts/sauce",
        )

        # test exceptions

        # error messages
        requests_error_message = "text2vec-contextionary vector was not retrieved."
        unexpected_exception_error_message = "text2vec-contextionary vector"

        ## test UnexpectedStatusCodeException
        contextionary = Contextionary(mock_connection_func("get", status_code=404))
        with self.assertRaises(UnexpectedStatusCodeException) as error:
            contextionary.get_concept_vector("Palantir")
        check_startswith_error_message(self, error, unexpected_exception_error_message)

        ## test requests error
        contextionary = Contextionary(
            mock_connection_func("get", side_effect=RequestsConnectionError("Test!"))
        )
        with self.assertRaises(RequestsConnectionError) as error:
            contextionary.get_concept_vector("Palantir")
        check_error_message(self, error, requests_error_message)
